!---------------------------------------------------------------------------------------
! PREP_REGIONS: get_surfinf
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> get_surfinf
!!
!! Purpose:    Calculate total surface influence from flexpart grid_time files.
!!
!! Interface:
!!
!!    Inputs
!!             files    -  files data structure
!!             recs     -  receptor IDs for each observation
!!             obstimes -  observation time stamps in julian days
!!             avetimes -  observation averaging time period in julian days
!!             surfinf  -  total surface influence (zero matrix on input)
!!
!!    Externals
!!             caldate
!!             read_grid
!!
!---------------------------------------------------------------------------------------

subroutine get_surfinf(files, config, recs, obstimes, avetimes, surfinf)

  use mod_var
  use mod_dates
  use mod_settings
  use mod_flexpart
  use mod_ncdf

  implicit none

  type (files_t),                     intent (in)     :: files
  type (config_t),                    intent (in)     :: config
  real(kind=8), dimension(maxobs),    intent (in)     :: obstimes
  real(kind=8), dimension(maxobs),    intent (in)     :: avetimes
  character(4), dimension(maxobs),    intent (in)     :: recs
  real, dimension(nxregrid,nyregrid), intent (in out) :: surfinf

  character(len=max_path_len)                         :: path_flexrec, filename, filedates
  character(len=max_name_len)                         :: varname
  character(len=6)                                    :: adate
  character(len=8)                                    :: areldate, areltime
  logical                                             :: lexist
  integer                                             :: ierr
  integer                                             :: jjjjmmdd, hhmiss, month, prevmonth
  integer                                             :: i, ix1, ix2, jy1, jy2

  real, dimension(nxgrid,nygrid,maxngrid)             :: grid       ! flexpart flux sensitivity (or footprints)
  real, dimension(nxgrid,nygrid,maxngrid)             :: grid_tmp   ! flexpart flux sensitivity (or footprints), temporary to later calculate average
  real(kind=8), dimension(maxngrid)                   :: gtime      ! flux sensitivity time stamp (julian days)
  real(kind=8), dimension(maxngrid)                   :: gtime_tmp  ! flux sensitivity time stamp (julian days) of first footprint per observation

  real                                                :: bndx, bndy, delx, dely
  integer                                             :: numx, numy, xshift, new_grid_time, standard_grid_time
  integer                                             :: numpoint, release_nr, nr_footprints, ngrid
  integer                                             :: ibdate, ibtime
  real(kind=8), dimension(maxpoint,2)                 :: releases

  ! indices to regional domain
  ! --------------------------

  ix1 = minloc(glon, dim=1, mask=abs(glon-dx/2.-rllx).eq.minval(abs(glon-dx/2.-rllx)))
  ix2 = minloc(glon, dim=1, mask=abs(glon+dx/2.-rllx-nxregrid*rdx).eq.minval(abs(glon+dx/2.-rllx-nxregrid*rdx)))
  jy1 = minloc(glat, dim=1, mask=abs(glat-dy/2.-rlly).eq.minval(abs(glat-dy/2.-rlly)))
  jy2 = minloc(glat, dim=1, mask=abs(glat+dy/2.-rlly-nyregrid*rdy).eq.minval(abs(glat+dy/2.-rlly-nyregrid*rdy)))

  print*, 'get_surfinf: glon(ix1), glon(ix2), glat(jy1), glat(jy2) = ',glon(ix1), glon(ix2), glat(jy1), glat(jy2)

  ! loop over obs times
  ! -------------------

  prevmonth = 0
  grid(:,:,:) = 0.

  do i = 1, nobs

    ! reset values
    gtime_tmp(:) = 0d0
    grid_tmp(:,:,:) = 0.
    nr_footprints = 0

    ! define file name
    call caldate(obstimes(i), jjjjmmdd, hhmiss)
    write(adate,'(I6.6)') jjjjmmdd/100
    write(areldate,'(I8.8)') jjjjmmdd
    write(areltime,'(I6.6)') hhmiss
    month = jjjjmmdd/100

    print*, 'Get footprints for observation nr:', i, recs(i), obstimes(i), jjjjmmdd, hhmiss
    path_flexrec = trim(files%path_flexpart)//trim(recs(i))//'/'//trim(adate)//'/'
    filename = trim(path_flexrec)//'header'

    ! get header and release information related to current observation location 
    ! (if not read and same for previous observation)
    if ( i.eq.1 .or. month.ne.prevmonth ) then
      inquire ( file=trim(filename),exist=lexist )
      if ( lexist ) then
        print*, 'Get header:', i, ' '//trim(path_flexrec)//'header'
        call read_header(filename, numpoint, ibdate, ibtime, releases, &
                           numx, numy, bndx, bndy, delx, dely, xshift)
      else
        print*, 'WARNING: cannot find ',filename
        go to 10
      endif
    else
      if(trim(recs(i)).ne.trim(recs(i-1)))then
        inquire ( file=trim(filename),exist=lexist )
        if ( lexist ) then
          print*, 'Get header:', i, ' '//trim(path_flexrec)//'header'
          call read_header(filename, numpoint, ibdate, ibtime, releases, &
                             numx, numy, bndx, bndy, delx, dely, xshift)
        else
          print*, 'WARNING: cannot find ',filename
          go to 10
        endif
      endif
    endif

    ! find the first release number corresponding to the current observation
    release_nr=1
    do while ( int(releases(release_nr,1)*1e5,kind=8).lt.int(obstimes(i)*1e5,kind=8) .and. release_nr.lt.numpoint )
      release_nr = release_nr+1
    end do

    ! define first filename
    call caldate(releases(release_nr,1), jjjjmmdd, hhmiss)
    ! round seconds
    hhmiss = (hhmiss/100)*100
    write(areldate,'(I8.8)') jjjjmmdd
    write(areltime,'(I6.6)') hhmiss
    if (.not.config%nested) filename = trim(path_flexrec)//'grid_time_'//trim(areldate)//trim(areltime)//'_001'
    if (config%nested) filename = trim(path_flexrec)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001'
    filedates = trim(path_flexrec)//'dates'

    ! read flexpart footprint
    inquire ( file=trim(filename),exist=lexist )
    if ( lexist ) then
      call read_grid(filename, filedates, obstimes(i), nxgrid, nygrid, nxshift, ngrid, grid, gtime)
      gtime_tmp = gtime  ! Save the output date of each field in footprint grid
      grid_tmp = grid    ! Save the footprint to temporary grid for summation
      nr_footprints = 1  ! Start counting number of footprints to be averaged   
     ! current release_nr was already read, start with next
      release_nr = release_nr+1
      do while ( int(releases(release_nr,2)*1e5,kind=8).le.int((obstimes(i)+avetimes(i))*1e5,kind=8) .and. release_nr.lt.numpoint )
        call caldate(releases(release_nr,1), jjjjmmdd, hhmiss)
        write(adate,'(I6.6)') jjjjmmdd/100
        write(areldate,'(I8.8)') jjjjmmdd
        hhmiss=int(hhmiss/100)*100
        write(areltime,'(I6.6)') hhmiss
        path_flexrec = trim(files%path_flexpart)//trim(recs(i))//'/'//trim(adate)//'/'
        if (.not.config%nested) filename = trim(path_flexrec)//'grid_time_'//trim(areldate)//trim(areltime)//'_001'
        if (config%nested) filename = trim(path_flexrec)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001'
        filedates = trim(path_flexrec)//'dates'
        call read_grid(filename, filedates, obstimes(i), nxgrid, nygrid, nxshift, ngrid, grid, gtime)
        nr_footprints=nr_footprints+1
        release_nr = release_nr+1
        ! sum the grid with previous footprints
        ! check for corresponding output date in gtime
        do standard_grid_time = 1, maxngrid
          do new_grid_time = 1, maxngrid
            if ( gtime(new_grid_time) .eq. gtime_tmp(standard_grid_time)) then
              grid_tmp(:,:,standard_grid_time) = grid_tmp(:,:,standard_grid_time)+grid(:,:,new_grid_time)
            endif
          end do
        end do
      end do ! finished summing footprints to get average
      print*, 'Total nr footprints for this observation:', nr_footprints
      grid = grid_tmp/real(nr_footprints)
      ! convert s.m3/kg to s.m2/kg
      grid = grid/outheight(1)
      ! add to total surface influence
      surfinf(:,:) = surfinf(:,:) + sum(grid(ix1:ix2,jy1:jy2,:),dim=3)
      prevmonth = month
    else
      ! file doesn't exist
      print*, 'WARNING: cannot find '//trim(filename)
      go to 10
    endif 

10  continue
  
  end do

  ! write surfinf to ncdf file
  filename = trim(files%path_output)//'surfinf.nc'
  varname = 'surfinf'
  call write_ncdf(files, filename, varname, surfinf)

end subroutine get_surfinf

