!---------------------------------------------------------------------------------------
! FLEXINVERT: retrieval
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> retrieval
!!
!! Purpose:    Interface to the optimization routines.
!!
!! Details:    Calculates the prior cost and gradient. If it is an optimization
!!             run, then optimizes the state vector and calculates the posterior
!!             cost (see also mod_analytic.f90 and congrad.f90).  
!!
!! Interface:
!!
!!    Inputs
!!             files   -  file data structure
!!             config  -  configuration settings data structure
!!             fluxes  -  flux data structure
!!             obs     -  observation data structure
!!             states  -  state vector data structure
!!             covar   -  error covariance data structure
!!
!!    Externals
!!             simulate
!!             congrad
!!             analytic
!!             prod_icov
!!             m1qn3_interface
!!   
!---------------------------------------------------------------------------------------

subroutine retrieval(files, config, fluxes, obs, states, covar)

  use mod_var
  use mod_settings
  use mod_fluxes
  use mod_obs
  use mod_states
  use mod_covar
  use mod_transform
  use mod_analytic

  implicit none

  type (files_t),  intent (in)      :: files
  type (config_t), intent (in)      :: config
  type (fluxes_t), intent (in)      :: fluxes
  type (obs_t),    intent (in out)  :: obs
  type (states_t), intent (in out)  :: states
  type (covar_t),  intent (in out)  :: covar

  real, dimension(nvar)             :: grad_p, grad_o, icovp 
  real(kind=8), dimension(nvar,1)   :: grad, chi_ad, phi, ones, chi, chi1 
  real                              :: cost_p, cost_o, cost
  integer                           :: iter
  integer                           :: ierr, n

  ! initialization
  ! --------------

  iter = 0
  grad_p(:) = 0.
  grad_o(:) = 0.
  grad(:,1) = 0d0
  cost_o = 0.
  cost_p = 0.
  ones(:,1) = 1d0

  if ( config%method.eq.'analytic' ) then
    ! allocate transport matrix
    allocate( hmat(nobs,nvar), stat=ierr ) 
    if (ierr.ne.0 ) then
      write(logid,*) 'ERROR retrieval: not enough memory' 
      stop
    endif
    hmat(:,:) = 0.
  else
    ! dummy variable
    allocate( hmat(1,1), stat=ierr )
    hmat(:,:) = 0.
  endif

  ! initial cost and gradient
  ! -------------------------

  ! gradient state space
  if ( config%lognormal ) then
    ! lognormal (note p is log transformed variable)
    ! J'p = ones + B^(-1)(p - p0)
    grad_p = ones(:,1) + prod_icov(config, covar, states%px - states%px0)
  else
    ! normal
    ! J'p = B^(-1)(p - p0)
    grad_p = prod_icov(config, covar, states%px - states%px0)
  endif

  ! cost state space
  if ( config%lognormal ) then
    ! lognormal (note p is log transformed variable)
    ! 2Jp = 2*sum(p) + (p - p0)^TB^(-1)(p - p0)
    icovp = prod_icov(config, covar, states%px - states%px0)
    cost_p = 2.*sum(states%px) + dot_product( states%px - states%px0, icovp )
  else
    ! normal
    ! 2Jp = (p - p0)^TB^(-1)(p - p0) 
    cost_p = dot_product( states%px - states%px0, grad_p )
  endif

  ! gradient observation space
  ! J'o = H^(T)R^(-1)(H(p) - y)
  if ( (config%method.eq.'congrad'.or.config%method.eq.'m1qn3').and.config%restart.eq.1 ) then
    open(100,file=trim(files%path_output)//'grad_00.txt',action='read',status='old',iostat=ierr)
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR retrieval: cannot open grad_00.txt'
      stop
    endif
    do n = 1, nvar
      read(100,*) grad_o(n)
    end do
    close(100)
    if ( config%method.eq.'m1qn3' ) then
      ! read cost 
      open(100,file=trim(files%path_output)//'cost.txt',action='read',status='old',iostat=ierr)
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR retrieval: cannot open cost.txt'
        stop
      endif
      read(100,*) cost
      close(100)
      write(logid,*) 'Cost read for iteration 0: ',cost
    endif
  else
    call simulate(iter, files, config, fluxes, obs, states, grad_o, cost_o)
    cost = 0.5*(cost_p + cost_o)
    write(logid,*) 'Initial cost 2Jp = ',cost_p
    write(logid,*) 'Initial cost 2Jo = ',cost_o
    write(logid,*) 'Initial cost: ', cost
  endif

  if (config%run_mode.eq.0) then
    write(logid,*) 'Finished forward model run'
    go to 10
  endif

  ! optimization 
  ! ------------

  if ( config%method.eq.'congrad' ) then
    ! total gradient
    ! J' = J'p + J'o 
    grad(:,1) = dble(grad_p + grad_o)
    ! transform grad to chi space
    chi_ad = phi2chi_ad(config, covar, grad)
    call congrad(iter, chi_ad, files, config, fluxes, obs, states, covar)
    iter = 99
    call simulate(iter, files, config, fluxes, obs, states, grad_o, cost_o)
  else if ( config%method.eq.'m1qn3' ) then
    ! total gradient
    ! J' = J'p + J'o
    grad(:,1) = dble(grad_p + grad_o)
    ! transform grad to chi space
    chi_ad = phi2chi_ad(config, covar, grad)
    call m1qn3_interface(iter, chi_ad, files, config, fluxes, obs, states, covar, cost)
    iter = 99
    call simulate(iter, files, config, fluxes, obs, states, grad_o, cost_o)   
  else if ( config%method.eq.'analytic' ) then
    call analytic(files, config, obs, states, covar, cost_o)
  else
    write(logid,*) 'ERROR: unknown retrieval method'
    stop
  endif
  
  ! cost state space
  if ( config%lognormal ) then
    ! lognormal (note p is log transformed variable)
    ! 2Jp = 2*sum(p) + (p - p0)^TB^(-1)(p - p0)
    icovp = prod_icov(config, covar, states%px - states%px0)
    cost_p = 2.*sum(states%px) + dot_product( states%px - states%px0, icovp )
  else
    ! normal
    ! 2Jp = (p - p0)^TB^(-1)(p - p0) 
    grad_p = prod_icov(config, covar, states%px - states%px0)
    cost_p = dot_product( states%px - states%px0, real(grad_p,kind=4) )
  endif

  write(logid,*) 'Final cost 2Jp = ',cost_p
  write(logid,*) 'Final cost 2Jo = ',cost_o

  ! total cost
  ! J = Jp + Jo 
  cost = 0.5*(cost_p + cost_o)
  write(logid,*) 'Final cost: ', cost

  write(logid,*) 'Finished inverse model run'

10 continue

  deallocate( hmat )

end subroutine retrieval


